/****************************************************************************
 * This file is part of the program mcmcalgn for joint sampling alignments  *
 * and mutation parameter values for two given DNA-sequences.               *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <climits>  

using namespace std;

#include "eingabe.hh"
#include "dnalign.hh"

double addlog(double a, double b) {
    // liefert log(exp(a)+exp(b))
    if (a>b) return a+log(1+exp(b-a));
    else return b+log(1+exp(a-b));
}

double addlog(double a, double b, double c) {
    // liefert log(exp(a)+exp(b)+exp(c))
    return addlog(addlog(a,b),c);
}

double maximum(double a, double b) {
    return ( a>b ? a : b );
}

double maximum(double a, double b, double c) {
    return maximum(maximum(a,b),c);
}

double rand_exp(double x) {
    /* gibt exp(x)-verteilte ZVE zurueck.
     * (exp-verteilt mit Erwartungswert 1/x)
     */
    double R=rand()/(RAND_MAX+1.0);
    if ( R>0.99 ) return -log(R)/x;
    else return -log(0.99)/x+rand_exp(x);
}

//+++++++++++++++++++++++++++++++++++++++++++
//             class dna_with_gaps          +
//+++++++++++++++++++++++++++++++++++++++++++

// private:

// public:

dna_with_gaps::dna_with_gaps(){
} 

dna_with_gaps::dna_with_gaps(ifstream& datei)
{
    char B;
    while(datei.get(B)) {
	switch (B) {
	case 'A':
	case 'a':
	    push_back('A');
	    break;
	case 'C':
	case 'c':
	    push_back('C');
	    break;
	case 'G':
	case 'g':
	    push_back('G');
	    break;
	case 'T':
	case 't':
	    push_back('T');
	    break;
	case '_':
	case '-':
	    push_back('_');
	    break;
	case '\n':
	case ' ':
	    break;
	default:
	    cerr << "Fehler in Eingabe-Datei: "<< B 
		 << " Bitte nur A,C,G,T !" << endl;
	}
    }
}

dna_with_gaps::dna_with_gaps(eingabe_parameter ep, int i) {
    
    ifstream datei;
    char B;
    switch (i) {
    case 1 :
	datei.open(ep.dnadat1.c_str());
	if (!datei) {
	    cerr << "cannot open file " << ep.dnadat1 << endl;
	    exit(EXIT_FAILURE);
	}
	break;
    case 2 :
	datei.open(ep.dnadat2.c_str());
	if (!datei) {
	    cerr << "cannot open file " << ep.dnadat2 << endl;
	    exit(EXIT_FAILURE);
	}
	break;
    default :
	cerr << "unsinniger Wert fuer i in "
	     << "dna_with_gaps::dna_with_gaps(eingabe_parameter ep, int i)\n";
	exit(1);
    } 
    while(datei.get(B)) {
	switch (B) {
	case 'A':
	case 'a':
	    push_back('A');
	    break;
	case 'C':
	case 'c':
	    push_back('C');
	    break;
	case 'G':
	case 'g':
	    push_back('G');
	    break;
	case 'T':
	case 't':
	    push_back('T');
	    break;
	case '_':
	case '-':
	    push_back('_');
	    break;
	case '\n':
	case ' ':
	    break;
	default:
	    cerr << "Fehler in Eingabe-Datei: "<< B 
		 << " Bitte nur A,C,G,T !" << endl;
	}
    }
    datei.close();   
}

dna_with_gaps::~dna_with_gaps(){}

int dna_with_gaps::laen() const{
    // gibt laenge zurueck
    return size();
}

ostream& operator<<(ostream& ostr,const dna_with_gaps& d) {
    for (int i=0; i<d.laen(); ++i){
	ostr << d[i];
    }
    return ostr;
}

ostream& operator<<(ostream& ostr,const zustand& z) {
    switch (z) {
    case PAARUNG : ostr << "PAARUNG"; break;
    case ENDE : ostr << "ENDE"; break;
    case A_GAP : ostr << "A_GAP"; break;
    case B_GAP : ostr << "B_GAP"; break;
    default: ostr << "NOSTATE";
    }
    return ostr;
}

//+++++++++++++++++++++++++++++++++++++++++++
//             class dna                    +
//+++++++++++++++++++++++++++++++++++++++++++

// private:

// public:

dna::dna(ifstream& datei)
{
    char B;
    while(datei.get(B)) {
	switch (B) {
	case 'A':
	case 'C':
	case 'G':
	case 'T':
	    push_back(B);
	    break;
	default:
	    cerr << "error in input file: "<< B 
		 << " only A,C,G, and T are allowed!" << endl;
	}
    }
}

dna::dna(const dna_with_gaps &dwg)
{
    for(int i=0; i<dwg.laen(); i++) {
	switch (dwg[i]) {
	case '_':
	    break;
	case 'A':
	case 'C':
	case 'G':
	case 'T':
	    push_back(dwg[i]);
	    break;
	default:
	    cerr << "Fehler beim Konstruieren einer DNA: Base " 
		 << dwg[i] << endl;
	}
    }
}

//+++++++++++++++++++++++++++++++++++++++++++
//        class alignment                   +
//+++++++++++++++++++++++++++++++++++++++++++

// private :

void alignment::set_max_wk() {
    /* Setzt das Alignment auf das wahrscheinlichste fuer die aktuellen 
     * Parameter
     */
    zustand nextstate=ENDE;
    vector<zustand> zust_vekt;
    maxwk_align_graph graph(*this);
    // cout << graph;
    int i=dna_A->laen(), j=dna_B->laen();
    while( i>0 || j>0 ) {
	nextstate=graph.maxwk_schritt(nextstate,i,j);
	switch (nextstate) {
	case PAARUNG : i--; j--; break;
	case A_GAP : j--; break;
	case B_GAP : i--; break;
	default: cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	}
	zust_vekt.push_back(nextstate);
    }
    int a=-1, b=-1;
    for( i=zust_vekt.size()-1; i>=0; i--) {
	switch(zust_vekt[i]) {
	case PAARUNG : 
	    a++; 
	    b++; 
	    dAB[a]=b;
	    dBA[b]=a; 
	    break;
	case A_GAP : 
	    b++; 
	    dBA[b]=a;
	    break;
	case B_GAP : 
	    a++;
	    dAB[a]=b;
	    break;
	default: 
	    cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	}
    }
}

double alignment::get_ln_wk(double s1, double l1, double m1) const {
    /* liefert den ln der Wkeit des aktuellen Alignments fuer die uebergebenen 
     * Parameter.
     */
    double log_uebwk[4][5];   // Wkeiten der Uebergaenge
    int anz_ueberg[4][5]      // Anzahl der einzelnen Uebergaenge
	={  {0,0,0,0,0},
	    {0,0,0,0,0},
	    {0,0,0,0,0},
	    {0,0,0,0,0} };
    const double beta=(1-exp((l1-m1)))/(m1-l1*exp((l1-m1)));
    log_uebwk[PAARUNG][B_GAP]=log((1-l1*beta)*l1/m1*(1-exp(-m1)));
    log_uebwk[PAARUNG][A_GAP]=log(l1*beta);
    log_uebwk[PAARUNG][PAARUNG]=log((1-l1*beta)*l1/m1*exp(-m1));
    log_uebwk[PAARUNG][ENDE]=log((1-l1/m1)*(1-l1*beta));
    log_uebwk[B_GAP][B_GAP]=log(l1*beta);
    log_uebwk[B_GAP][A_GAP]=log((1-exp(-m1)-m1*beta)/(1-exp(-m1)));
    log_uebwk[B_GAP][PAARUNG]=log(l1*beta*exp(-m1)/(1-exp(-m1)));
    log_uebwk[B_GAP][ENDE]=log((1-l1/m1)*m1*beta/(1-exp(-m1)));
    log_uebwk[A_GAP][B_GAP]=log((1-l1*beta)*l1/m1*(1-exp(-m1)));
    log_uebwk[A_GAP][A_GAP]=log(l1*beta);
    log_uebwk[A_GAP][PAARUNG]=log((1-l1*beta)*l1/m1*exp(-m1));
    log_uebwk[A_GAP][ENDE]=log((1-l1/m1)*(1-l1*beta));
    log_uebwk[START][B_GAP]=log((1-l1*beta)*l1/m1*(1-exp(-m1)));
    log_uebwk[START][A_GAP]=log(l1*beta);
    log_uebwk[START][PAARUNG]=log((1-l1*beta)*l1/m1*exp(-m1));
    log_uebwk[START][ENDE]=log((1-l1/m1)*(1-l1*beta));
    int anz_matches=0,anz_mismatches=0;
    zaehle_ueberg(anz_ueberg,&anz_matches, &anz_mismatches,
		  -1,-1,dna_A->laen(),dna_B->laen());
    double summe=0;
    for( int i=0; i<4; i++ ){
	summe+=log_uebwk[i][0]*anz_ueberg[i][0];
	summe+=log_uebwk[i][1]*anz_ueberg[i][1]+log(0.25)*anz_ueberg[i][1];
	summe+=log_uebwk[i][2]*anz_ueberg[i][2]+log(0.25)*anz_ueberg[i][2];
	summe+=log_uebwk[i][4]*anz_ueberg[i][4];
    }
    summe+=log(0.25*(1-3*0.25*(1-exp(-s1))))*anz_matches
	+log(0.25*0.25*(1-exp(-s1)))*anz_mismatches;
    return summe;
}


void alignment::zaehle_ueberg(int anz_ueberg[4][5],
			      int * p_anz_matches,int * p_anz_mismatches,
			      int von_A,int von_B,int bis_A,int bis_B)
    const {
    /* Berechnet die Anzahlen der verschiedenen Uebergaenge zwischen den
     * Zustaenden und zaehlt die matches und mismatches im aktuellen
     * Alignment im angegebenen Bereich 
     * und schreibt die Ergebnisse in das Feld anz_ueberg bzw. nach
     * *p_anz_matches und *p_anz_mismatches.
     */
    zustand oldstate=get_zust(von_A,von_B), zust;
    int a=von_A, b=von_B;
    for (int i=0;i<4;i++)
	for (int j=0;j<5;j++)
	    anz_ueberg[i][j]=0;
    *p_anz_matches=0;
    *p_anz_mismatches=0;
    if ( a!=-1 && b!= -1 && dBA[b]!=a && dAB[a]!=b || 
	 ( bis_A != dna_A->laen() && dAB[bis_A]!=bis_B
	   && bis_B != dna_B->laen() && dBA[bis_B]!=bis_A ) )
	cerr << "Fehler bei den Bereichsgrenzen in alignment::zaehle_ueberg\n"
	     << setw(4) << a << setw(4) << b 
	     <<setw(4) << bis_A << setw(4) << bis_B << endl 
	     << setw(4) << dBA[b] << setw(4) << dAB[a] 
	     <<setw(4) << dna_A->laen() << setw(4) << dna_B->laen() 
	     << endl << endl ;

    do {
	zust=get_zust(a,b);
	switch (zust) {
	case PAARUNG :
	    if((*dna_A)[a]==(*dna_B)[b])
		++*p_anz_matches;
	    else
		++*p_anz_mismatches;
	case START :
	case A_GAP :
	case B_GAP :
	case ENDE :
	    ++(anz_ueberg[oldstate][zust]);
	    break;
	default:	 
	    cerr << "Inkompatibilitaet zwischen a und b in"
		 << " alignment::zaehle_ueberg(..)!!!\n"
		 << "a=" << a << " dAB[a]=" << dAB[a]
		 << " b="  << b << " dBA[b]=" << dBA[b]<< " " <<endl;
	    exit(1);
	}
	oldstate=iter_zust(&a,&b);
    } while(zust!=ENDE && a<=bis_A && b<=bis_B );
}

zustand alignment::get_zust(int a,int b) const {
    /* Liefert den zu a und b assoziierten Alignment-Zustand
     */
    if (a==-1) {
	if (b==-1) 
	    return START;
	if (dBA[b]==-1) 
	    return A_GAP;
	return NOSTATE;
    }
    if (a==dna_A->laen()) {
	if (b==dna_B->laen())
	    return ENDE;
	return NOSTATE;
    }
    if (b==-1) {
	if (dAB[a]==-1) 
	    return B_GAP;
	return NOSTATE;
    }
    if (b==dna_B->laen()) {
	return NOSTATE;
    }
    if (b==dAB[a]){
	if (a==dBA[b])
	    return PAARUNG;
	if (dBA[b]<a)
	    return B_GAP;
	return NOSTATE;
    }
    if (dBA[b]==a && dAB[a]<b )
	return A_GAP;
    return NOSTATE;
}

zustand alignment::iter_zust(int* pa,int* pb) const {
    /* Liefert den zu *pa und *pb assoziierten Alignment-Zustand und 
     * setzt (*pa,*pb) einen Zustand weiter.
   */
    zustand zust;
    zust=get_zust(*pa,*pb);
    switch (zust) {
    case START:
    case PAARUNG:  
    case A_GAP: 
    case B_GAP:
	if ( *pa==dna_A->laen()-1 ) {
	    ++*pb;
	    if ( *pb==dna_B->laen() ) 
		++*pa;
	}
	else {
	    if ( *pb==dna_B->laen()-1 )
		++*pa;
	    else {
		if ( dAB[*pa+1]==*pb )
		    ++*pa;
		else {
		    ++*pb;
		    if ( dBA[*pb]!=*pa ) 
			++*pa;
		}
	    }
	}
	break;
    case ENDE:
	break;
    default:
	cerr << *pa << " und " << *pb << " ergeben keinen Zustand"
	     << " in alignment::iter_zust\n";
	exit(1);
    }
    return zust;
}

void alignment::teilsample(int von_A, int von_B, int bis_A, int bis_B) 
{
    /* sample ein neues Teilalignment von (von_A,von_B) 
     * bis (bis_A,bis_B). Dabei bleiben die Zustaende der von
     * Start- und Endpunkt erhalten.
     */
    
    zustand nextstate=get_zust(bis_A,bis_B);
    vector<zustand> zust_vekt;
    int a=bis_A, b=bis_B;
    int anz_A_GAP=0,anz_B_GAP=0;
    while (iter_zust(&a,&b)==A_GAP) ++anz_A_GAP;
    if (anz_A_GAP==0) {
	a=bis_A;
	b=bis_B;
	while (iter_zust(&a,&b)==B_GAP) ++anz_B_GAP;
    }
    wk_align_graph graph(*this,von_A,von_B,bis_A,bis_B);
    // cout << graph;
    // double ln_wk_alt=get_ln_wk();
    int i=0, j=0;
    switch (nextstate) {
    case ENDE :
	i=dna_A->laen();
	j=dna_B->laen();
	break;
    case PAARUNG :
	i=bis_A;
	j=bis_B;
	break;
    case A_GAP :
	i=bis_A+1;
	j=bis_B;
	break;
    case B_GAP :
	i=bis_A;
	j=bis_B+1;
	break;
    default:
	cerr << "unmoeglicher \"nextstate\" in teilsample(...)\n";
    }
    //   cout << endl << "A:"<< von_A << '-' << bis_A << ' ' 
    //        << "B:" << von_B << '-' << bis_B<< endl;
    while( i>von_A+1 || j>von_B+1 ) {
	nextstate=graph.zufallsschritt(nextstate,i,j);
	switch (nextstate) {
	case PAARUNG : i--; j--; break;
	case A_GAP : j--; break;
	case B_GAP : i--; break;
	default: cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	}
	if ( i<von_A || j<von_B ) 
	    cerr << "Bereich verlassen in  alignment::teilsample(...)\n";
	zust_vekt.push_back(nextstate);
	// cout << i << " " << j << " " << nextstate << endl;
    }
    a=von_A;
    b=von_B;
    for( i=zust_vekt.size()-1; i>=0; i--) {
	switch(zust_vekt[i]) {
	case PAARUNG : 
	    a++; 
	    b++;  
	    dAB[a]=b;
	    dBA[b]=a;
	    break;
	case A_GAP :
	    b++;  
	    dBA[b]=a;
	    break;
	case B_GAP :
	    a++;
	    dAB[a]=b; 
	    break;
	case ENDE :
	    break;
	default: 
	    cerr << "unmoeglicher \"nextstate\" in teilsample!!!\n";
	}
	//cout << a << " " << b << " " <<  setw(10) << zust_vekt[i] 
	//	 << ' ' << (*dna_A)[a] << ' ' <<  (*dna_B)[b] <<endl;
    }
    for( i=1; i<anz_A_GAP; i++){
	dBA[bis_B+i]=dBA[bis_B];
	//cout << "dBA[" << bis_B+i << "]=" << dBA[bis_B] << endl;
    }
    for( i=1; i<anz_B_GAP; i++){
	dAB[bis_A+i]=dAB[bis_A];
	//cout <<"dAB[" << bis_A+i << "]=" << dAB[bis_A] << endl;
    }
    // if (get_ln_wk()<ln_wk_alt-100){
    //   cerr << "Drastische Reduziereung der Wahrscheinlichkeit in teilsample" 
    //   << setw(4) << von_A<< setw(4) << von_B
    //   << setw(4) << bis_A<< setw(4) << bis_B<<endl
    //   << *this << endl << graph << endl;
    //   }
}

double alignment::quot_prior_s(double s1,double s2) 
{
    /* liefert den Quotient zwischen der Dichte der a-priori-Verteilung
     * von s an den Stellen s1 und s2.
     */
    return exp(s2-s1);
}

double alignment::quot_prior_lambda(double lambda_1,double lambda_2){
    /* liefert den Quotient zwischen der Dichte der a-priori-Verteilung
     * von lambda an den Stellen lambda_1 und lambda_2.
     */
    return exp(lambda_2-lambda_1);
}

double alignment::quot_prior_mu(double mu_1,double mu_2){
    /* liefert den Quotient zwischen der Dichte der a-priori-Verteilung
     * von mu an den Stellen mu_1 und mu_2.
     */
    return exp(mu_2-mu_1);
}


// public:

alignment::alignment(){
    /* Gibt eine Fehlermeldung
     */
    cerr << "Vorsicht! Ein leeres Alignment wurde konstruiert!\n";
}

alignment::alignment(const dna &d_1,const dna &d_2)
    : dAB(d_1.laen()), dBA(d_2.laen())
    /* Initialisiert mit Alignment maximaler Wahrscheinlichkeit unter
     * Verwendung von Defaultwerten fuer die Mutationsparameter
     */
{ 
    dna_A=&d_1;
    dna_B=&d_2;
    cerr << "ACHTUNG: Es wird ein Alignment mit s=0.1, lambda=0.1 "
	 << "und mu=0.11 konstruiert.\n";
    s=0.1;
    lambda=0.1;
    mu=0.11;
    set_max_wk();
}

alignment::alignment(const dna &d_1,const dna &d_2,
		     const dna_with_gaps &d_1wg,const dna_with_gaps &d_2wg,
		     double s0,double l,double m)
    : dAB(d_1.laen()), dBA(d_2.laen())
    /* Initialisiert mit dem durch die Gaps in Sequenzen vorgegebenen Alignment
     */
{ 
    dna_A=&d_1;
    dna_B=&d_2;
    s=s0;
    lambda=l;
    mu=m;
    if(d_1wg.laen()!=d_2wg.laen()) {
	cerr << "dna_with_gap's have different lenghts!" << endl;
	exit(1);
    }
    int a=-1,b=-1;
    for(int i=0; i<d_1wg.laen(); i++){
	if(d_1wg[i]=='_') {
	    if(d_2wg[i]!='_') {
		b++;
		dBA[b]=a;
	    }
	}
	else {
	    if(d_2wg[i]=='_') {
		a++;
		dAB[a]=b;
	    }
	    else {
		a++;
		b++;
		dAB[a]=b;
		dBA[b]=a;
	    } 
	}
    }
}

alignment::alignment(const alignment& align)
    /* Copy-Kostruktor
     */
{
    dAB=align.dAB;
    dBA=align.dBA;
    dna_A=align.dna_A;
    dna_B=align.dna_B;
    s=align.s;
    lambda=align.lambda;
    mu=align.mu;
}

alignment::alignment(const dna &d_1,const dna &d_2,
		     double s0,double l,double m)
    /* Initialisiert das Alignment als das mit maximaler 
     * Wahrscheinlichkeit
     */
    : dAB(d_1.laen()), dBA(d_2.laen())
{ 
    dna_A=&d_1;
    dna_B=&d_2;
    s=s0;
    lambda=l;
    mu=m;
    set_max_wk();
}

alignment::alignment(const dna &d_1,const dna &d_2,
	      const dna_with_gaps &d_1wg,const dna_with_gaps &d_2wg,
	      const eingabe_parameter & ep)
    /* Konstruiert das Alignment gemaess den Eingabeparametern in eg
     */ 
    : dAB(d_1.laen()), dBA(d_2.laen())
{
    int a=-1,b=-1;
    switch (ep.start_align) {
    case 0 : 
	dna_A=&d_1;
	dna_B=&d_2;
	s=ep.s;
	lambda=ep.lambda;
	mu=ep.mu;
	if(d_1wg.laen()!=d_2wg.laen()) {
	    cerr << "dna_with_gap's have different lengths!" << endl;
	    exit(1);
	}
	for(int i=0; i<d_1wg.laen(); i++){
	    if(d_1wg[i]=='_') {
		if(d_2wg[i]!='_') {
		    b++;
		    dBA[b]=a;
		}
	    }
	    else {
		if(d_2wg[i]=='_') {
		    a++;
		    dAB[a]=b;
		}
		else {
		    a++;
		    b++;
		    dAB[a]=b;
		    dBA[b]=a;
		} 
	    }
	}
	break;
    case 1 :
	dna_A=&d_1;
	dna_B=&d_2;
	s=ep.s;
	lambda=ep.lambda;
	mu=ep.mu;
	set_max_wk();
	sample();
	break;
    case 2 : 
	dna_A=&d_1;
	dna_B=&d_2;
	s=ep.s;
	lambda=ep.lambda;
	mu=ep.mu;
	set_max_wk();
	break;
    default :
	cerr << "unknown type of initial alignment" << endl;
	exit(EXIT_FAILURE);
    }
}

void alignment::sample()
    /* sample das Alignment
     */
{
    teilsample(-1,-1,dna_A->laen(),dna_B->laen());
//   int a[3][4],anz_matches,anz_mismatches;
//   zaehle_ueberg(a,&anz_matches, &anz_mismatches, 
// 		  0,0,dna_A->laen(),dna_B->laen());
//   for( int i=0;i<3;i++){
//     for (int j=0 ; j<4;j++)
//       cout << setw(4) << a[i][j];
//     cout << endl;
//   }
//   cout << anz_matches << " " << anz_mismatches << endl;
} 

void alignment::teilsample(int von_A, int bis_A) {
    /* sample ein neues Teilalignment von dna_A[von_A] 
     * bis dna_A[bis_A].
   */
    teilsample(von_A,dAB[von_A],bis_A,dAB[bis_A]);
}

void alignment::teilsample(int laenge) {
    /* sample ein neues Teilalignment zu einem zufaelligen Abschnitt der 
     * Laenge laenge. 
   */
    if (laenge<1) cerr << "unsinniger Parameter in alignment::teilsample(...)\n";
    if (laenge>=dna_A->laen() ) sample();
    else {
	int R=rand()%(dna_A->laen()+laenge);
	if ( R >= dna_A->laen() )
	    teilsample( -1,-1, laenge, dAB[laenge]);
	else {
	    if ( R+laenge >= dna_A->laen() ) 
		teilsample( dna_A->laen()-laenge, dAB[dna_A->laen()-laenge],
			    dna_A->laen(),dna_B->laen());
	    else
		teilsample( R,R+laenge);
	}
    }
}

void alignment::param_resamp() {
    /* Fuehrt einen Metropolis-Hastings-Schritt fuer s, lambda und mu
     * bedingt auf das aktuelle Alignment durch.
     */
    double new_s=rand_exp(1/s),new_lambda=rand_exp(1/lambda),
	g=rand_exp((mu-lambda)/lambda), new_mu=new_lambda*(g+1)/g;
    double R=(rand()+1.0)/(RAND_MAX+1.0);
    double quotient=exp(get_ln_wk(new_s,new_lambda,new_mu)-get_ln_wk())
	*quot_prior_s(new_s,s)
	*quot_prior_lambda(new_lambda,lambda)
	*quot_prior_mu(new_mu,mu)
	*s/new_s*exp(new_s/s-s/new_s)
	*lambda/new_lambda*exp(new_lambda/lambda-lambda/new_lambda)
	*lambda*(new_mu-new_lambda)/((mu-lambda)*new_lambda)
	*exp(new_lambda*(mu-lambda)/((new_mu-new_lambda)*lambda)
	     -lambda*(new_mu-new_lambda)/((mu-lambda)*new_lambda));
    if ( R < quotient ) {
	s=new_s;
	lambda=new_lambda;
	mu=new_mu;
    }
}

void alignment::print_params(ostream& ostr) {
    /* gibt s, lambda und ln(W'keit) in den ostr aus
     */
    ostr << setw(15) << s  
	 << setw(15) << lambda 
	 << setw(15) << get_ln_wk() << "  ";
}

void alignment::print_path(ostream& ostr) {
    /* gibt das Alignment Pfad in ostr aus:
     * d,v,h stehen fuer diagonal, vertikal und horizontal
     */    
    int i=0,a=-1,b=-1;
    string pfad;
    iter_zust(&a,&b);
    zustand zust;
    do {
	zust=get_zust(a,b);
	switch (zust) {
	case PAARUNG :
	    pfad+='d';
	    break;
	case A_GAP :
	    pfad+='h';
	    break;
	case B_GAP :
	    pfad+='v';
	    break;
	case ENDE :
	    break;
	default:
	    cerr << "incompatibility between a and b" <<endl;
	    exit(1);
	}
	i++;
	zust=iter_zust(&a,&b);
    } while(zust!=ENDE);
    ostr << pfad;
}

double alignment::get_ln_wk() const {
    /* liefert den ln der Wkeit des aktuellen Alignments fuer die aktuellen 
     * Parameter.
     */
    return get_ln_wk(s,lambda,mu); 
}

ostream& operator<<(ostream& ostr, const alignment& algn) {
    int i=0,a=-1,b=-1,anz_matches=0,anz_mismatches=0,anz_A_gaps=0,anz_B_gaps=0;
    string wort_A, wort_B;
    algn.iter_zust(&a,&b);
    zustand zust;
    do {
	zust=algn.get_zust(a,b);
	switch (zust) {
	case PAARUNG :
	    wort_A+=(*algn.dna_A)[a];
	    wort_B+=(*algn.dna_B)[b];
	    if ((*algn.dna_A)[a]==(*algn.dna_B)[b])
		anz_matches++;
	    else
		anz_mismatches++;
	    break;
	case A_GAP :
	    wort_A+='_';
	    wort_B+=(*algn.dna_B)[b];
	    anz_A_gaps++;
	    break;
	case B_GAP :
	    wort_A+=(*algn.dna_A)[a];
	    wort_B+='_';
	    anz_B_gaps++;
	    break;
	case ENDE :
	    break;
	default:
	    ostr << wort_A << endl << wort_B << endl;
	    cerr << "incompatibility between a and b in"
		 << " ostream& operator<<(...)!!!\n"
		 << "a=" << a << " dAB[a]=" << algn.dAB[a]
		 << " b="  << b << " dBA[b]=" << algn.dBA[b]<< " " <<endl;
	    exit(1);
	}
	i++;
	if( i%70==0) {
	    ostr << wort_A << endl << wort_B << endl << endl;
	    wort_A="";
	    wort_B="";
	}
	zust=algn.iter_zust(&a,&b);
    } while(zust!=ENDE);
    ostr << wort_A << endl << wort_B << endl
	 << "s=" << algn.s << " lambda=" << algn.lambda 
	 << " mu=" << algn.mu << endl
	 << "matches: " << anz_matches << "\t mismatches: " << anz_mismatches << endl
	 << "gaps in 1st sequence : " << anz_A_gaps 
	 << "\tgaps in 2nd seqence: " <<  anz_B_gaps << endl
	 << "logarithmus naturalis of probability of this alignment : "
	 << algn.get_ln_wk() << endl;
//   int az=-1, bz=-1;
//   zustand zust=algn.iter_zust(&az,&bz);
//   while (zust!=ENDE) {
//     zust=algn.iter_zust(&az,&bz);
//     cout << zust << ",";
//   }
    return ostr << endl;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++
//             class align_graph                     +
//++++++++++++++++++++++++++++++++++++++++++++++++++++

//  private :

void align_graph::berechne_log_uebwk() {
    /* berechnet die log_uebwk[][] fuer die jeweiligen Werte von
     * lambda und mu
     */
    const double beta=(1-exp((lambda-mu)))/(mu-lambda*exp((lambda-mu)));
    log_uebwk[PAARUNG][B_GAP]=log((1-lambda*beta)*lambda/mu*(1-exp(-mu)));
    log_uebwk[PAARUNG][A_GAP]=log(lambda*beta);
    log_uebwk[PAARUNG][PAARUNG]=log((1-lambda*beta)*lambda/mu*exp(-mu));
    log_uebwk[PAARUNG][ENDE]=log((1-lambda/mu)*(1-lambda*beta));
    log_uebwk[B_GAP][B_GAP]=log(lambda*beta);
    log_uebwk[B_GAP][A_GAP]=log((1-exp(-mu)-mu*beta)/(1-exp(-mu)));
    log_uebwk[B_GAP][PAARUNG]=log(lambda*beta*exp(-mu)/(1-exp(-mu)));
    log_uebwk[B_GAP][ENDE]=log((1-lambda/mu)*mu*beta/(1-exp(-mu)));
    log_uebwk[A_GAP][B_GAP]=log((1-lambda*beta)*lambda/mu*(1-exp(-mu)));
    log_uebwk[A_GAP][A_GAP]=log(lambda*beta);
    log_uebwk[A_GAP][PAARUNG]=log((1-lambda*beta)*lambda/mu*exp(-mu));
    log_uebwk[A_GAP][ENDE]=log((1-lambda/mu)*(1-lambda*beta));  
    log_uebwk[START][B_GAP]=log((1-lambda*beta)*lambda/mu*(1-exp(-mu)));
    log_uebwk[START][A_GAP]=log(lambda*beta);
    log_uebwk[START][PAARUNG]=log((1-lambda*beta)*lambda/mu*exp(-mu));
    log_uebwk[START][ENDE]=log((1-lambda/mu)*(1-lambda*beta));
    log_uebwk[ENDE][B_GAP]=-999999999.9;
    log_uebwk[ENDE][A_GAP]=-999999999.9;
    log_uebwk[ENDE][PAARUNG]=-999999999.9;
    log_uebwk[ENDE][ENDE]=-999999999.9;
//    for (int i=0; i<4; i++){
//      for (int j=0; j<4; j++)
//        cout << setw(10) << exp(log_uebwk[i][j]);
//      cout << endl;
//    }
}

//  public :

align_graph::align_graph(const alignment& algn,
			 int v_A,int v_B,int b_A,int b_B)
    /* Konstruiere einen Alignment-Graphen fuer das Teilalignment zwischen
   * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
   * von algn
   */
    : A_gap_kante(b_A-v_A+2,vector<double>(b_B-v_B+1,999999.9)) ,
      B_gap_kante(b_A-v_A+1,vector<double>(b_B-v_B+2,999999.9)) ,
      paar_kante(b_A-v_A+1,vector<double>(b_B-v_B+1,999999.9))
{
    sub=0.25*(1-exp(-algn.s));
    lambda=algn.lambda;
    mu=algn.mu;
    von_A=v_A; bis_A=b_A; von_B=v_B; bis_B=b_B;
}

align_graph::align_graph(const alignment& algn) :
    A_gap_kante(algn.get_dna_A()->laen()+1, 
		vector<double>(algn.get_dna_B()->laen(),999999.9)) ,
    B_gap_kante(algn.get_dna_A()->laen(),
		vector<double>(algn.get_dna_B()->laen()+1,999999.9)) ,
    paar_kante(algn.get_dna_A()->laen(),
	       vector<double>(algn.get_dna_B()->laen(),999999.9)  )
    
{
    /* Konstruiere einen Alignment-Graphen fuer das Alignment
     */
    sub=0.25*(1-exp(-algn.s));
    lambda=algn.lambda;
    mu=algn.mu;
    von_A=-1; bis_A=algn.get_dna_A()->laen();
    von_B=-1; bis_B=algn.get_dna_B()->laen();
}

align_graph::align_graph()
{
    cerr << "FEHLER: ein leerer align_graph wurde konstruiert!!!\n";
}

ostream& operator<<(ostream& ostr, const align_graph& ag) {
    for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	ostr << setw(30) << i+ag.von_A+1;
    ostr<<endl;
    for(int j=0; j<=ag.bis_B-ag.von_B-2; j++) {
	ostr << setw(5) << " ";
	for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	    ostr <<  setw(30) << ag.B_gap_kante[i][j];
	ostr<<endl;
	ostr << setw(5) << j+ag.von_B+1;
	for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	    ostr <<  setw(15) << ag.A_gap_kante[i][j]
		 <<  setw(15) << ag.paar_kante[i][j];
	ostr << setw(15) << ag.A_gap_kante[ag.bis_A-ag.von_A-1][j];
	ostr<<endl;
    }
    ostr << setw(5) << " ";
    for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	ostr << setw(30) << ag.B_gap_kante[i][ag.bis_B-ag.von_B-1];
    ostr<<endl;
    return ostr;
}

//+++++++++++++++++++++++++++++++++++++++++++++
//       class   wk_align_graph               +
//+++++++++++++++++++++++++++++++++++++++++++++

//  private :

void wk_align_graph::beschrifte(const alignment& algn) 
    /* Beschriftet die Kanten des align_graph fuer den angegebenen 
   * Bereich gemaess den DNA-Staengen und den Parametern von algn
   */
{  
    const dna &dna_A=*(algn.get_dna_A());
    const dna &dna_B=*(algn.get_dna_B());
    zustand firststate = algn.get_zust(von_A,von_B), 
	laststate  = algn.get_zust(bis_A,bis_B);
    int v_A=von_A+1,v_B=von_B+1,b_A=0,b_B=0;
    switch (laststate) {
    case ENDE :
	b_A=algn.dna_A->laen()-1;
	b_B=algn.dna_B->laen()-1;
	break;
    case PAARUNG :
	b_A=bis_A-1;
	b_B=bis_B-1;
	break;
    case A_GAP :
	b_A=bis_A;
	b_B=bis_B-1;
	break;
    case B_GAP :
	b_A=bis_A-1;
	b_B=bis_B;
	break;
    default:
	cerr << "komischer laststate in wk_align_graph::beschrifte(...)\n";
    } 
    double c, c1, c2, cx;
    // Zur W'keit fuer Substitution verwandte Logarithmen
    int i,j;
    c=log(0.25);
    c1=log(0.25*(1-3*sub));
    c2=log(0.25*sub);
    berechne_log_uebwk();
    A_gap_kante[0][0]=log_uebwk[firststate][A_GAP]+c;
    B_gap_kante[0][0]=log_uebwk[firststate][B_GAP]+c;
    paar_kante[0][0]=log_uebwk[firststate][PAARUNG]
	+ ( dna_A[v_A]==dna_B[v_B] ? c1 : c2 );
    for(i=1;i<=b_A-v_A;i++) {
	A_gap_kante[i][0]=B_gap_kante[i-1][0]+log_uebwk[B_GAP][A_GAP]+c;
	B_gap_kante[i][0]=B_gap_kante[i-1][0]+log_uebwk[B_GAP][B_GAP]+c;
	paar_kante[i][0]=B_gap_kante[i-1][0]+log_uebwk[B_GAP][PAARUNG]
	    + ( dna_A[i+v_A]==dna_B[v_B] ? c1 : c2 );
    }
    A_gap_kante[b_A-v_A+1][0]=B_gap_kante[b_A-v_A][0]
	+log_uebwk[B_GAP][A_GAP]+c;
    for(j=1;j<=b_B-v_B;j++) {
	B_gap_kante[0][j]=A_gap_kante[0][j-1]+log_uebwk[A_GAP][B_GAP]+c;
	A_gap_kante[0][j]=A_gap_kante[0][j-1]+log_uebwk[A_GAP][A_GAP]+c;
	paar_kante[0][j]=A_gap_kante[0][j-1]+log_uebwk[A_GAP][PAARUNG]
	    + ( dna_A[v_A]==dna_B[j+v_B] ? c1 : c2 );
    }
    B_gap_kante[0][b_B-v_B+1]=(A_gap_kante[0][b_B-v_B]
			       +log_uebwk[A_GAP][B_GAP]+c);
    for(i=1;i<=b_A-v_A;i++) {
	for(j=1;j<=b_B-v_B;j++) {
	    A_gap_kante[i][j]
		=addlog(B_gap_kante[i-1][j]+log_uebwk[B_GAP][A_GAP]+c,
			paar_kante[i-1][j-1]+log_uebwk[PAARUNG][A_GAP]+c,
			A_gap_kante[i][j-1]+log_uebwk[A_GAP][A_GAP]+c);
	    B_gap_kante[i][j]
		=addlog(B_gap_kante[i-1][j]+log_uebwk[B_GAP][B_GAP]+c,
			paar_kante[i-1][j-1]+log_uebwk[PAARUNG][B_GAP]+c,
			A_gap_kante[i][j-1]+log_uebwk[A_GAP][B_GAP]+c);
	    cx=( dna_A[i+v_A]==dna_B[j+v_B] ? c1 : c2 );
	    paar_kante[i][j]
		=addlog(B_gap_kante[i-1][j]+log_uebwk[B_GAP][PAARUNG]+cx,
			paar_kante[i-1][j-1]+log_uebwk[PAARUNG][PAARUNG]+cx,
			A_gap_kante[i][j-1]+log_uebwk[A_GAP][PAARUNG]+cx);
	}
    }
    for(i=1;i<=b_A-v_A;i++) 
	B_gap_kante[i][b_B-v_B+1]
	    =addlog(B_gap_kante[i-1][b_B-v_B+1]+log_uebwk[B_GAP][B_GAP]+c,
		    paar_kante[i-1][b_B-v_B]+log_uebwk[PAARUNG][B_GAP]+c,
		    A_gap_kante[i][b_B-v_B]+log_uebwk[A_GAP][B_GAP]+c);
    for(j=1;j<=b_B-v_B;j++)
	A_gap_kante[b_A-v_A+1][j]
	    =addlog(B_gap_kante[b_A-v_A][j]+log_uebwk[B_GAP][A_GAP]+c,
		    paar_kante[b_A-v_A][j-1]+log_uebwk[PAARUNG][A_GAP]+c,
		    A_gap_kante[b_A-v_A+1][j-1]+log_uebwk[A_GAP][A_GAP]+c);
}

zustand wk_align_graph::zufallsschritt(zustand nextstep, int a, int b) const
    /* Gibt einen zufaelligen Zustand aus, der dem Schritt hin zu Knoten
     * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
     * es von (a,b) aus mit einem Schritt von Zustand nextstep weitergeht.
     */
{
    a-=von_A+1; b-=von_B+1;
    if ( a<0 || b<0 ) 
	cerr << "Fehler: zu kleine Parameter in align_graph::zufallsschritt!!!\n";
    if (a<=0) {
	if (b<=0) cerr << "Fehler: a,b<=0 in  align_graph::zufallsschritt !!!\n";
	//cout << A_GAP <<endl;
	return A_GAP;
    }
    if (b==0) {
	//cout << B_GAP << endl;
	return B_GAP;
    }
    double R=rand()/(RAND_MAX+1.0);
    double l_Paar=paar_kante[a-1][b-1]+log_uebwk[PAARUNG][nextstep],
	l_A_GAP=A_gap_kante[a][b-1]+log_uebwk[A_GAP][nextstep],
	l_B_GAP=B_gap_kante[a-1][b]+log_uebwk[B_GAP][nextstep], 
	l_summe=addlog(l_Paar,l_A_GAP,l_B_GAP),
	Wk_A_gap=exp(l_A_GAP-l_summe),
	Wk_B_gap=exp(l_B_GAP-l_summe);
    //   cout <<  setw(12) << l_Paar << setw(12) << l_A_GAP<< setw(12) << l_B_GAP
    //        << setw(12) << Wk_A_gap << setw(12)  << Wk_B_gap << setw(4) 
    //        << a << setw(4) << b << setw(12) << R << endl;
    if ( R < Wk_A_gap ) {
	//cout << A_GAP << endl;
	return A_GAP;
    }
    if ( R < Wk_A_gap+Wk_B_gap ) {
	//cout << B_GAP << endl;
	return B_GAP;
    }
    //cout << PAARUNG << endl;
    return PAARUNG;
}  

// public :
wk_align_graph::wk_align_graph(const alignment& algn,
			       int v_A,int v_B,int b_A,int b_B)
    /* Konstruiere einen wk-Alignment-Graphen fuer das Teilalignment zwischen
   * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
   * von algn
   */
    :align_graph(algn,v_A,v_B,b_A,b_B)
{
    beschrifte(algn);
}

//++++++++++++++++++++++++++++++++++++++++++++
//    class maxwk_align_graph                +
//++++++++++++++++++++++++++++++++++++++++++++

// private:

void maxwk_align_graph::beschrifte(const alignment& algn) 
    /* Beschriftet die Kanten des align_graph 
   * gemaess den DNA-Staengen und den Parametern von algn
   */
{  
    const dna &dna_A=*(algn.get_dna_A());
    const dna &dna_B=*(algn.get_dna_B());
    double c, c1, c2, cx;
    // Zur W'keit fuer Substitution verwandte Logarithmen
    int i,j;
    c=log(0.25);
    c1=log(0.25*(1-3*sub));
    c2=log(0.25*sub);
    berechne_log_uebwk();
    A_gap_kante[0][0]=log_uebwk[PAARUNG][A_GAP]+c;
    B_gap_kante[0][0]=log_uebwk[PAARUNG][B_GAP]+c;
    paar_kante[0][0]=log_uebwk[PAARUNG][PAARUNG]
	+ ( dna_A[0]==dna_B[0] ? c1 : c2 );
    for(i=1;i<=dna_A.laen()-1;i++) {
	A_gap_kante[i][0]=B_gap_kante[i-1][0]+log_uebwk[B_GAP][A_GAP]+c;
	B_gap_kante[i][0]=B_gap_kante[i-1][0]+log_uebwk[B_GAP][B_GAP]+c;
	paar_kante[i][0]=B_gap_kante[i-1][0]+log_uebwk[B_GAP][PAARUNG]
	    + ( dna_A[i+0]==dna_B[0] ? c1 : c2 );
    }
    A_gap_kante[dna_A.laen()][0]=B_gap_kante[dna_A.laen()-1][0]
	+log_uebwk[B_GAP][A_GAP]+c;
    for(j=1;j<=dna_B.laen()-1;j++) {
	B_gap_kante[0][j]=A_gap_kante[0][j-1]+log_uebwk[A_GAP][B_GAP]+c;
	A_gap_kante[0][j]=A_gap_kante[0][j-1]+log_uebwk[A_GAP][A_GAP]+c;
	paar_kante[0][j]=A_gap_kante[0][j-1]+log_uebwk[A_GAP][PAARUNG]
	    + ( dna_A[0]==dna_B[j+0] ? c1 : c2 );
    }
    B_gap_kante[0][dna_B.laen()]=(A_gap_kante[0][dna_B.laen()-1]
				  +log_uebwk[A_GAP][B_GAP]+c);
    for(i=1;i<=dna_A.laen()-1;i++) {
	for(j=1;j<=dna_B.laen()-1;j++) {
	    A_gap_kante[i][j]
		=maximum(B_gap_kante[i-1][j]+log_uebwk[B_GAP][A_GAP]+c,
			 paar_kante[i-1][j-1]+log_uebwk[PAARUNG][A_GAP]+c,
			 A_gap_kante[i][j-1]+log_uebwk[A_GAP][A_GAP]+c);
	    B_gap_kante[i][j]
		=maximum(B_gap_kante[i-1][j]+log_uebwk[B_GAP][B_GAP]+c,
			 paar_kante[i-1][j-1]+log_uebwk[PAARUNG][B_GAP]+c,
			 A_gap_kante[i][j-1]+log_uebwk[A_GAP][B_GAP]+c);
	    cx=( dna_A[i]==dna_B[j] ? c1 : c2 );
	    paar_kante[i][j]
		=maximum(B_gap_kante[i-1][j]+log_uebwk[B_GAP][PAARUNG]+cx,
			 paar_kante[i-1][j-1]+log_uebwk[PAARUNG][PAARUNG]+cx,
			 A_gap_kante[i][j-1]+log_uebwk[A_GAP][PAARUNG]+cx);
	}
    }
    for(i=1;i<=dna_A.laen()-1;i++) 
	B_gap_kante[i][dna_B.laen()]
	    =maximum(B_gap_kante[i-1][dna_B.laen()]+log_uebwk[B_GAP][B_GAP]+c,
		     paar_kante[i-1][dna_B.laen()-1]+log_uebwk[PAARUNG][B_GAP]+c,
		     A_gap_kante[i][dna_B.laen()-1]+log_uebwk[A_GAP][B_GAP]+c);
    for(j=1;j<=dna_B.laen()-1;j++)
	A_gap_kante[dna_A.laen()][j]
	    =maximum(B_gap_kante[dna_A.laen()-1][j]+log_uebwk[B_GAP][A_GAP]+c,
		     paar_kante[dna_A.laen()-1][j-1]+log_uebwk[PAARUNG][A_GAP]+c,
		     A_gap_kante[dna_A.laen()][j-1]+log_uebwk[A_GAP][A_GAP]+c);
}

// public :


maxwk_align_graph::maxwk_align_graph(const alignment& algn)
    : align_graph(algn)
{
    /* Konstruiere einen Alignment-Graphen fuer das Alignment
     */
    beschrifte(algn);
}


zustand maxwk_align_graph::maxwk_schritt(zustand nextstep, int a, int b) const
    /* Gibt den wahrscheinlichsten Zustand aus, der dem Schritt hin zu Knoten
   * (a,b) entspricht, und zwar bezueglich der bedingten Verteilung, dass
   * es von (a,b) aus mit einem Schritt von Zustand nextstep weitergeht.
   */
{
    //cout << nextstep << endl;
    if ( a<0 || b<0 ) 
	cerr << "Fehler: zu kleine Parameter in align_graph::zufallsschritt!!!\n";
    if (a==0) {
	if (b==0) cerr << "Fehler: a=b=0 in  align_graph::zufallsschritt !!!\n";
	return A_GAP;
    }
    if (b==0) return B_GAP;
    double l_Paar=paar_kante[a-1][b-1]+log_uebwk[PAARUNG][nextstep],
	l_A_GAP=A_gap_kante[a][b-1]+log_uebwk[A_GAP][nextstep],
	l_B_GAP=B_gap_kante[a-1][b]+log_uebwk[B_GAP][nextstep];
    if ( l_B_GAP < l_A_GAP ) { 
	if( l_Paar <  l_A_GAP )
	    return A_GAP;
	else
	    return PAARUNG;
    }
    else {
	if( l_Paar <  l_B_GAP )
	    return B_GAP;
	else
	    return PAARUNG;
    }
}  

