/****************************************************************************
 * This file is part of the program mcmcalgn for joint sampling alignments  *
 * and mutation parameter values for two given DNA-sequences.               *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

class dna_with_gaps : public vector <char> {
public:
    dna_with_gaps();
    dna_with_gaps(ifstream& datei);
    dna_with_gaps(eingabe_parameter ep, int i);
    ~dna_with_gaps();
    int laen() const; // gibt laenge zurueck
};

class dna : public dna_with_gaps {
public:
    dna(ifstream& datei);
    dna(const dna_with_gaps &dwg);
};

ostream& operator<<(ostream& ostr, const dna_with_gaps& d);

class alignment;

enum zustand {PAARUNG=0,A_GAP=1,B_GAP=2,START=3,ENDE=4,NOSTATE=5};
//--------------------------------------------------------------------
class align_graph {
    friend ostream& operator<<(ostream& ostr, const align_graph& ag);
protected:
    double sub,lambda,mu;
    int von_A, von_B, bis_A, bis_B; 
    vector< vector<double> > A_gap_kante;
    vector< vector<double> > B_gap_kante;
    vector< vector<double> > paar_kante; 
    double log_uebwk[5][5]; // Uebergangswkeiten zwischen den Zustaenden
    void berechne_log_uebwk()
	/* berechnet die log_uebwk[][] fuer die jeweiligen Werte von
	 * delta, lambda und mu
	 */;
public:
    align_graph(const alignment& algn,int von_A,int von_B,int bis_A,int bis_B)
	/* Konstruiere einen Alignment-Graphen fuer das Teilalignment zwischen
	 * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
	 * von algn
	 */;
    align_graph(const alignment& algn)
	/* Konstruiere einen Alignment-Graphen fuer das Alignment
	 */;
    align_graph()
	/* Fuehrt zu Fehlermeldung
	 */;
};
//--------------------------------------------------------------------
class wk_align_graph : public  align_graph {
private:
    void beschrifte(const alignment& algn) 
	/* Beschriftet die Kanten des maxwk_align_graph fuer den angegebenen 
	 * Bereich gemaess den DNA-Staengen und den Parametern von algn
	 */;
public:
    wk_align_graph(const alignment& algn,
		   int von_A,int von_B,int bis_A,int bis_B)
	/* Konstruiere einen maxwk_Alignment-Graphen fuer das Teilalignment
	 * zwischen
	 * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
	 * von algn
	 */;
    zustand zufallsschritt(zustand s, int a, int b) const
	/* Gibt einen zufaelligen Zustand aus, der dem Schritt hin zu Knoten
	 * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
	 * es von (a,b) aus mit einem Schritt von Zustand s weitergeht.
	 */;
};
//--------------------------------------------------------------------
class maxwk_align_graph : public  align_graph {
private: 
    void beschrifte(const alignment& algn) 
	/* Beschriftet die Kanten des maxwk_align_graph fuer den angegebenen 
	 * Bereich gemaess den DNA-Staengen und den Parametern von algn
	 */;
public:
    maxwk_align_graph(const alignment& algn)
	/* Konstruiere einen maxwk_Alignment-Graphen fuer das Alignment
	 */;
    zustand maxwk_schritt(zustand s, int a, int b) const
	/* Gibt den wahrscheilichsten Zustand aus, der dem Schritt hin zu Knoten
	 * (a,b) entspricht, und zwar bezueglich der bedingten Verteilung, dass
	 * es von (a,b) aus mit einem Schritt von Zustand s weitergeht.
	 */;
};

ostream& operator<<(ostream& ostr, const align_graph& ag);
//--------------------------------------------------------------------
class alignment {
    friend class align_graph;
    friend class wk_align_graph;
    friend class maxwk_align_graph;
    friend ostream& operator<<(ostream& ostr, const alignment& algn);
private:
    const dna *dna_A, *dna_B;
    double s, lambda, mu;
    vector<int> dAB,dBA;
    void set_max_wk()
	/* Setzt das Alignment auf das wahrscheinlichste fuer die aktuellen 
	 * Parameter
	 */;
    double get_ln_wk(double s, double lambda, double mu) const 
	/* liefert den ln der Wkeit des aktuellen Alignments
	 * fuer die uebergebenen Parameter.
	 */;
    void zaehle_ueberg(int anz_ueberg[4][5],
		       int * p_anz_matches,int * p_anz_mismatches,
		       int von_A,int von_B,int bis_A,int bis_B) const
	/* Berechnet die Anzahlen der verschiedenen Uebergaenge zwischen den
	 * Zustaenden und zaehlt die matches und mismatches im aktuellen
	 * Alignment im angegebenen Bereich 
	 * und schreibt die Ergebnisse in das Feld anz_ueberg bzw. nach
	 * *p_anz_matches und *p_anz_mismatches.
	 */;
    zustand get_zust(int a,int b) const
	/* Liefert den zu a und b assoziierten Alignment-Zustand
	 */; 
    zustand iter_zust(int* pa,int* pb) const
	/* Liefert den zu *pa und *pb assoziierten Alignment-Zustand und 
	 * setzt (*pa,*pb) einen Zustand weiter.
	 */;
    void teilsample(int von_A, int von_B, int bis_A, int bis_B) 
	/* sample ein neues Teilalignment von (von_A,von_B) 
	 * bis (bis_A,bis_B). Dabei bleiben die Zustaende der von
	 * Start- und Endpunkt erhalten.
	 */;
    double quot_prior_s(double s1,double s2)
	/* liefert den Quotient zwischen der Dichte der a-priori-Verteilung
	 * von s an den Stellen s1 und s2.
	 */;
    double quot_prior_lambda(double lambda_1,double lambda_2)
	/* liefert den Quotient zwischen der Dichte der a-priori-Verteilung
	 * von lambda an den Stellen lambda_1 und lambda_2.
	 */;
    double quot_prior_mu(double mu_1,double mu_2)
	/* liefert den Quotient zwischen der Dichte der a-priori-Verteilung
	 * von mu an den Stellen mu_1 und mu_2.
	 */;
public:
    alignment()
	/* Gibt eine Fehlermeldung
	 */;
    alignment(const dna &d_1,const dna &d_2)
	/* Initialisiert mit Alignment maximaler Wahrscheinlichkeit unter
	 * Verwendung von Defaultwerten fuer die Mutationsparameter
	 */;
    alignment(const dna &d_1,const dna &d_2,
	      const dna_with_gaps &d_1wg,const dna_with_gaps &d_2wg,
	      double s, double lambda, double mu)
	/* Initialisiert mit dem durch die Gaps in Sequenzen vorgegebenen
	 * Alignment
	 */;
    alignment(const alignment& align)
	/* Copy-Kostruktor
	 */;
    alignment(const dna &d_1,const dna &d_2, double s, double lambda, double mu)
	/* Initialisiert das Alignment als das mit maximaler 
	 * Wahrscheinlichkeit bei den uebergebenen Mutationsparametern
	 */;
    alignment(const dna &d_1,const dna &d_2,
	      const dna_with_gaps &d_1wg,const dna_with_gaps &d_2wg,
	      const eingabe_parameter & ep)
	/* Konstruiert das Alignment gemaess den Eingabeparametern in eg
	 */;
    const dna* get_dna_A() const {return dna_A;};
    const dna* get_dna_B() const {return dna_B;};
    void sample()
	/* sample das Alignment
	 */;
    void teilsample(int von_A, int bis_A)
	/* sample ein neues Teilalignment von dna_A[von_A] 
	 * bis dna_A[bis_A].
	 */;
    void teilsample(int laenge)
	/* sample ein neues Teilalignment zu einem zufaelligen Abschnitt der 
	 * Laenge laenge. 
	 */;
    void param_resamp()
	/* Fuehrt einen Metropolis-Hastings-Schritt fuer s
	 * lambda und mu bedingt auf das aktuelle Alignment durch.
	 */;  
    void print_params(ostream& ostr)
	/* gibt s, lambda und ln(W'keit) in den ostr aus
	 */;
    void alignment::print_path(ostream& ostr) 
	/* gibt das Alignment Pfad in ostr aus:
	 * d,v,h stehen fuer diagonal, vertikal und horizontal
	 */;
    double get_ln_wk() const
	/* liefert den ln der Wkeit des aktuellen Alignments fuer die aktuellen 
	 * Parameter.
	 */;
};

ostream& operator<<(ostream& ostr, const alignment& algn);
